@@ -1,15 +1,19 @@
+import logging
 import sys
 import time
 from contextlib import contextmanager
-from queue import LifoQueue
+from queue import Empty, LifoQueue
 from threading import Thread
 
 from django.apps import apps
 from django.db.models import signals as model_signals
 
-from . import filebrowser
-from . import update_lock
-from linkcheck.models import Url, Link
+from linkcheck.models import Link, Url
+
+from . import filebrowser, update_lock
+from .linkcheck_settings import MAX_URL_LENGTH
+
+logger = logging.getLogger(__name__)
 
 
 tasks_queue = LifoQueue()
@@ -17,11 +21,25 @@ worker_running = False
 tests_running = len(sys.argv) > 1 and sys.argv[1] == 'test' or sys.argv[0].endswith('runtests.py')
 
 
-def linkcheck_worker():
+def linkcheck_worker(block=True):
     global worker_running
     while tasks_queue.not_empty:
-        task = tasks_queue.get()
-        task['target'](*task['args'], **task['kwargs'])
+        try:
+            task = tasks_queue.get(block=block)
+        except Empty:
+            break
+        # An error in any task should not stop the worker from continuing with the queue
+        try:
+            task['target'](*task['args'], **task['kwargs'])
+        except Exception as e:
+            logger.exception(
+                "%s while running %s with args=%r and kwargs=%r: %s",
+                type(e).__name__,
+                task['target'].__name__,
+                task['args'],
+                task['kwargs'],
+                e
+            )
         tasks_queue.task_done()
     worker_running = False
 
@@ -61,7 +79,7 @@ def check_instance_links(sender, instance, **kwargs):
             new_links = []
             old_links = Link.objects.filter(content_type=content_type, object_id=instance.pk)
 
-            linklists = linklist_cls().get_linklist(extra_filter={'pk':instance.pk,})
+            linklists = linklist_cls().get_linklist(extra_filter={'pk': instance.pk})
 
             if not linklists:
                 # This object is no longer watched by linkcheck according to object_filter
@@ -77,10 +95,17 @@ def check_instance_links(sender, instance, **kwargs):
                 if url.startswith('#'):
                     internal_hash = url
                     url = instance.get_absolute_url() + url
+
+                if len(url) > MAX_URL_LENGTH:
+                    # We cannot handle url longer than MAX_URL_LENGTH at the moment
+                    logger.warning('URL exceeding max length will be skipped: %s', url)
+                    continue
+
                 u, created = Url.objects.get_or_create(url=url)
-                l, created = Link.objects.get_or_create(url=u, field=link[0], text=link[1], content_type=content_type, object_id=instance.pk)
+                l, created = Link.objects.get_or_create(
+                    url=u, field=link[0], text=link[1], content_type=content_type, object_id=instance.pk
+                )
                 new_links.append(l.id)
-                u.still_exists = True
                 if internal_hash:
                     setattr(u, '_internal_hash', internal_hash)
                     setattr(u, '_instance', instance)
