@@ -1,8 +1,10 @@
 from django.core.management.base import BaseCommand
 
+from linkcheck.linkcheck_settings import (
+    EXTERNAL_RECHECK_INTERVAL,
+    MAX_CHECKS_PER_RUN,
+)
 from linkcheck.utils import check_links
-from linkcheck.linkcheck_settings import EXTERNAL_RECHECK_INTERVAL
-from linkcheck.linkcheck_settings import MAX_CHECKS_PER_RUN
 
 
 class Command(BaseCommand):
@@ -10,12 +12,16 @@ class Command(BaseCommand):
     help = 'Check and record external link status'
 
     def add_arguments(self, parser):
-        parser.add_argument('-e', '--externalinterval', type=int,
+        parser.add_argument(
+            '-e', '--externalinterval', type=int,
             help='Specifies the length of time in minutes until external links are rechecked. '
-                 'Defaults to linkcheck_config setting')
-        parser.add_argument('-l', '--limit', type=int,
+                 'Defaults to linkcheck_config setting'
+        )
+        parser.add_argument(
+            '-l', '--limit', type=int,
             help='Specifies the maximum number (int) of links to be checked. '
-                 'Defaults to linkcheck_config setting.  Value less than 1 will check all')
+                 'Defaults to linkcheck_config setting.  Value less than 1 will check all'
+        )
 
     def handle(self, *args, **options):
         externalinterval = options['externalinterval'] or EXTERNAL_RECHECK_INTERVAL
@@ -27,4 +33,3 @@ class Command(BaseCommand):
 
         check_count = check_links(external_recheck_interval=externalinterval, limit=limit, check_internal=False)
         return "%s external URLs have been checked." % (check_count)
-
