@@ -1,6 +1,9 @@
 from django.core.management.base import BaseCommand
 
-from linkcheck.linkcheck_settings import EXTERNAL_RECHECK_INTERVAL, MAX_CHECKS_PER_RUN
+from linkcheck.linkcheck_settings import (
+    EXTERNAL_RECHECK_INTERVAL,
+    MAX_CHECKS_PER_RUN,
+)
 from linkcheck.utils import check_links
 
 
@@ -9,12 +12,16 @@ class Command(BaseCommand):
     help = 'Check and record internal and external link status'
 
     def add_arguments(self, parser):
-        parser.add_argument('-e', '--externalinterval', type=int,
+        parser.add_argument(
+            '-e', '--externalinterval', type=int,
             help='Specifies the length of time in minutes until external links are rechecked. '
-                 'Defaults to linkcheck_config setting')
-        parser.add_argument('-l', '--limit', type=int,
+                 'Defaults to linkcheck_config setting'
+        )
+        parser.add_argument(
+            '-l', '--limit', type=int,
             help='Specifies the maximum number (int) of links to be checked. '
-                 'Defaults to linkcheck_config setting.  Value less than 1 will check all')
+                 'Defaults to linkcheck_config setting.  Value less than 1 will check all'
+        )
 
     def handle(self, *args, **options):
         externalinterval = options['externalinterval'] or EXTERNAL_RECHECK_INTERVAL
