@@ -1,5 +1,10 @@
 import time
-from django.http import HttpResponse, HttpResponsePermanentRedirect, HttpResponseRedirect
+
+from django.http import (
+    HttpResponse,
+    HttpResponsePermanentRedirect,
+    HttpResponseRedirect,
+)
 
 
 def http_response(request, code):
@@ -22,3 +27,11 @@ def http_redirect_to_404(request):
 def timeout(request):
     time.sleep(2)
     return HttpResponse("")
+
+
+def http_response_with_anchor(request):
+    return HttpResponse("<html><body><h1 id='anchor'>Anchor</h1></body></html>")
+
+
+def http_redirect_to_anchor(request):
+    return HttpResponseRedirect("/http/anchor/")
