@@ -1,12 +1,8 @@
+import os
 from datetime import datetime, timedelta
 from io import StringIO
-from unittest import skipIf
-from urllib import request
-from urllib.error import HTTPError
-import os
-import re
+from unittest.mock import patch
 
-import django
 from django.apps import apps
 from django.conf import settings
 from django.contrib.auth.models import User
@@ -15,117 +11,117 @@ from django.test import LiveServerTestCase, TestCase
 from django.test.utils import override_settings
 from django.urls import reverse
 
+from linkcheck.linkcheck_settings import MAX_URL_LENGTH
 from linkcheck.listeners import (
-    enable_listeners, disable_listeners, register_listeners, unregister_listeners,
+    disable_listeners,
+    enable_listeners,
+    linkcheck_worker,
+    register_listeners,
+    tasks_queue,
+    unregister_listeners,
 )
 from linkcheck.models import Link, Url
 from linkcheck.views import get_jquery_min_js
 
-from .sampleapp.models import Author, Book
-
-
-#MOCK addinfurl
-class addinfourl():
-    """class to add info() and geturl(url=) methods to an open file."""
-
-    def __init__(self, url, code, msg):
-        self.headers = None
-        self.url = url
-        self.code = code
-        self.msg  = msg
-
-    def info(self):
-        return self.headers
-
-    def getcode(self):
-        return self.code
-
-    def geturl(self):
-        return self.url
-
-#
-# Mock Method so test can run independently
-#
-
-def mock_urlopen(url, data=None, **kwargs):
-    msg_dict = {'301': "Moved Permanently", '404': 'Not Found', '200': 'OK'}
-
-    code = '404'
-    msg  = msg_dict.get(code)
-
-    m = re.search("([0-9]*)$", url)
-    if m:
-        code = m.group(0)
-        msg  = msg_dict.get(code, 'Something Happened')
-        if code == "200":
-            return addinfourl(url, code, msg)
-
-    raise HTTPError(url, code, msg, None, None)
+from .sampleapp.models import Author, Book, Journal
 
 
 @override_settings(ROOT_URLCONF='linkcheck.tests.urls')
 class InternalCheckTestCase(TestCase):
 
-    def setUp(self):
-        #replace urllib2.urlopen with mock method
-        request.urlopen = mock_urlopen
-
     def test_internal_check_mailto(self):
-        uv = Url(url="mailto:nobody", still_exists=True)
+        uv = Url(url="mailto:nobody")
         uv.check_url()
         self.assertEqual(uv.status, None)
         self.assertEqual(uv.message, 'Email link (not automatically checked)')
 
     def test_internal_check_blank(self):
-        uv = Url(url="", still_exists=True)
+        uv = Url(url="")
         uv.check_url()
         self.assertEqual(uv.status, False)
         self.assertEqual(uv.message, 'Empty link')
 
-    def test_internal_check_anchor(self):
-        uv = Url(url="#some_anchor", still_exists=True)
+    def test_same_page_anchor(self):
+        uv = Url(url="#some_anchor")
         uv.check_url()
         self.assertEqual(uv.status, None)
         self.assertEqual(uv.message, 'Link to within the same page (not automatically checked)')
 
-    def test_internal_check_view_redirect(self):
-        uv = Url(url="/admin/linkcheck", still_exists=True)
+    def test_working_internal_anchor(self):
+        uv = Url(url="/http/anchor/#anchor")
         uv.check_url()
         self.assertEqual(uv.status, True)
-        self.assertIn(uv.message,
-            ['This link redirects: code %s (Working redirect)' % status for status in [301, 302]]
-        )
-        uv = Url(url="/http/brokenredirect/", still_exists=True)
+        self.assertEqual(uv.message, "Working internal link, working internal hash anchor")
+
+    @patch("linkcheck.models.TOLERATE_BROKEN_ANCHOR", False)
+    def test_broken_internal_anchor(self):
+        uv = Url(url="/http/anchor/#broken-anchor")
         uv.check_url()
         self.assertEqual(uv.status, False)
-        self.assertEqual(uv.message, 'This link redirects: code 302 (Broken redirect)')
+        self.assertEqual(uv.message, "Working internal link, broken internal hash anchor")
+
+    def test_broken_internal_anchor_tolerated(self):
+        uv = Url(url="/http/anchor/#broken-anchor")
+        uv.check_url()
+        self.assertEqual(uv.status, True)
+        self.assertEqual(uv.message, "Working internal link, broken internal hash anchor")
+
+    def test_redirect_working_internal_anchor(self):
+        uv = Url(url="/http/redirect_to_anchor/#anchor")
+        uv.check_url()
+        self.assertEqual(uv.status, True)
+        self.assertEqual(uv.message, "Working temporary redirect, working internal hash anchor")
+
+    @patch("linkcheck.models.TOLERATE_BROKEN_ANCHOR", False)
+    def test_redirect_broken_internal_anchor(self):
+        uv = Url(url="/http/redirect_to_anchor/#broken-anchor")
+        uv.check_url()
+        self.assertEqual(uv.status, False)
+        self.assertEqual(uv.message, "Working temporary redirect, broken internal hash anchor")
+
+    def test_redirect_broken_internal_anchor_tolerated(self):
+        uv = Url(url="/http/redirect_to_anchor/#broken-anchor")
+        uv.check_url()
+        self.assertEqual(uv.status, True)
+        self.assertEqual(uv.message, "Working temporary redirect, broken internal hash anchor")
+
+    def test_internal_check_view_redirect(self):
+        uv = Url(url="/admin/linkcheck")
+        uv.check_url()
+        self.assertEqual(uv.status, True)
+        self.assertEqual(uv.message, "Working temporary redirect")
+        uv = Url(url="/http/brokenredirect/")
+        uv.check_url()
+        self.assertEqual(uv.status, False)
+        self.assertEqual(uv.message, 'Broken temporary redirect')
 
     def test_internal_check_found(self):
-        uv = Url(url="/public/", still_exists=True)
+        uv = Url(url="/public/")
         uv.check_url()
         self.assertEqual(uv.status, True)
         self.assertEqual(uv.message, 'Working internal link')
+        self.assertEqual(uv.type, 'internal')
+
+    def test_internal_check_with_protocol(self):
+        # "localhost" is configured as SITE_DOMAIN in settings
+        uv = Url(url="http://localhost/public/")
+        uv.check_url()
+        self.assertEqual(uv.status, True)
+        self.assertEqual(uv.message, 'Working internal link')
+        self.assertEqual(uv.type, 'internal')
 
     def test_internal_check_broken_internal_link(self):
-        uv = Url(url="/broken/internal/link", still_exists=True)
+        uv = Url(url="/broken/internal/link")
         uv.check_url()
         self.assertEqual(uv.status, False)
         self.assertEqual(uv.message, 'Broken internal link')
 
     def test_internal_check_invalid_url(self):
-        uv = Url(url="invalid/url", still_exists=True)
+        uv = Url(url="invalid/url")
         uv.check_url()
         self.assertEqual(uv.status, False)
         self.assertEqual(uv.message, 'Invalid URL')
 
-    def test_same_page_anchor(self):
-        # TODO Make this test
-        pass
-        #uv = Url(url="#anchor", still_exists=True)
-        #uv.check_url()
-        #self.assertEqual(uv.status, None)
-        #self.assertEqual(uv.message, "")
-
 
 class InternalMediaCheckTestCase(TestCase):
     def setUp(self):
@@ -136,13 +132,13 @@ class InternalMediaCheckTestCase(TestCase):
         settings.MEDIA_ROOT = self.old_media_root
 
     def test_internal_check_media_missing(self):
-        uv = Url(url="/media/not_found", still_exists=True)
+        uv = Url(url="/media/not_found")
         uv.check_url()
         self.assertEqual(uv.status, False)
         self.assertEqual(uv.message, 'Missing Document')
 
     def test_internal_check_media_found(self):
-        uv = Url(url="/media/found", still_exists=True)
+        uv = Url(url="/media/found")
         uv.check_url()
         self.assertEqual(uv.status, True)
         self.assertEqual(uv.message, 'Working file link')
@@ -151,94 +147,163 @@ class InternalMediaCheckTestCase(TestCase):
         media_file = os.path.join(os.path.dirname(__file__), 'media', 'rückmeldung')
         open(media_file, 'a').close()
         self.addCleanup(os.remove, media_file)
-        uv = Url(url="/media/r%C3%BCckmeldung", still_exists=True)
+        uv = Url(url="/media/r%C3%BCckmeldung")
         uv.check_url()
         self.assertEqual(uv.status, True)
         self.assertEqual(uv.message, 'Working file link')
         # Also when the url is not encoded
-        uv = Url(url="/media/rückmeldung", still_exists=True)
+        uv = Url(url="/media/rückmeldung")
         uv.check_url()
         self.assertEqual(uv.status, True)
         self.assertEqual(uv.message, 'Working file link')
 
 
-# See https://code.djangoproject.com/ticket/29849 (fixed in Django 2.1+)
-@skipIf(django.VERSION[:2]==(2, 0), 'LiveServerTestCase is broken on Django 2.0.x')
 @override_settings(SITE_DOMAIN='example.com')
 class ExternalCheckTestCase(LiveServerTestCase):
     def test_external_check_200(self):
-        uv = Url(url="%s/http/200/" % self.live_server_url, still_exists=True)
+        uv = Url(url=f"{self.live_server_url}/http/200/")
         uv.check_url()
         self.assertEqual(uv.status, True)
         self.assertEqual(uv.message, '200 OK')
         self.assertEqual(uv.redirect_to, '')
 
+    def test_external_check_200_missing_cert(self):
+        uv = Url(url=f"{self.live_server_url.replace('http://', 'https://')}/http/200/")
+        uv.check_url()
+        self.assertEqual(uv.status, False)
+        self.assertEqual(uv.message, 'SSL Error: wrong version number')
+        self.assertEqual(uv.redirect_to, '')
+
     def test_external_check_200_utf8(self):
-        uv = Url(url="%s/http/200/r%%C3%%BCckmeldung/" % self.live_server_url, still_exists=True)
+        uv = Url(url=f"{self.live_server_url}/http/200/r%C3%BCckmeldung/")
         uv.check_url()
         self.assertEqual(uv.status, True)
         self.assertEqual(uv.message, '200 OK')
         # Also when the url is not encoded
-        uv = Url(url="%s/http/200/rückmeldung/" % self.live_server_url, still_exists=True)
+        uv = Url(url=f"{self.live_server_url}/http/200/rückmeldung/")
         uv.check_url()
         self.assertEqual(uv.status, True)
         self.assertEqual(uv.message, '200 OK')
 
     def test_external_check_301(self):
-        uv = Url(url="%s/http/301/" % self.live_server_url, still_exists=True)
+        uv = Url(url=f"{self.live_server_url}/http/301/")
         uv.check_url()
         self.assertEqual(uv.status, False)
-        self.assertEqual(uv.message.lower(), '301 moved permanently')
+        self.assertEqual(uv.message, '301 Moved Permanently')
         self.assertEqual(uv.redirect_to, '')
 
     def test_external_check_301_followed(self):
-        uv = Url(url="%s/http/redirect/301/" % self.live_server_url, still_exists=True)
+        uv = Url(url=f"{self.live_server_url}/http/redirect/301/")
         uv.check_url()
         self.assertEqual(uv.status, True)
         self.assertEqual(uv.message, '301 Moved Permanently')
-        self.assertEqual(uv.redirect_to, '%s/http/200/' % self.live_server_url)
+        self.assertEqual(uv.redirect_to, f'{self.live_server_url}/http/200/')
 
     def test_external_check_302_followed(self):
-        uv = Url(url="%s/http/redirect/302/" % self.live_server_url, still_exists=True)
+        uv = Url(url=f"{self.live_server_url}/http/redirect/302/")
         uv.check_url()
         self.assertEqual(uv.status, True)
         self.assertEqual(uv.message, '302 Found')
-        self.assertEqual(uv.redirect_to, '%s/http/200/' % self.live_server_url)
+        self.assertEqual(uv.redirect_to, f'{self.live_server_url}/http/200/')
 
     def test_external_check_404(self):
-        uv = Url(url="%s/whatever/" % self.live_server_url, still_exists=True)
+        uv = Url(url=f"{self.live_server_url}/whatever/")
         uv.check_url()
         self.assertEqual(uv.status, False)
         self.assertEqual(uv.message.lower(), '404 not found')
 
     def test_external_check_redirect_final_404(self):
-        uv = Url(url="%s/http/redirect_to_404/" % self.live_server_url, still_exists=True)
+        uv = Url(url=f"{self.live_server_url}/http/redirect_to_404/")
         uv.check_url()
         self.assertEqual(uv.status, False)
         self.assertEqual(uv.message.lower(), '404 not found')
 
     def test_external_check_get_only(self):
         # An URL that allows GET but not HEAD, linkcheck should fallback on GET.
-        uv = Url(url="%s/http/getonly/405/" % self.live_server_url, still_exists=True)
+        uv = Url(url=f"{self.live_server_url}/http/getonly/405/")
         uv.check_url()
         self.assertEqual(uv.status, True)
         self.assertEqual(uv.message, '200 OK')
         # Same test with other 40x error
-        uv = Url(url="%s/http/getonly/400/" % self.live_server_url, still_exists=True)
+        uv = Url(url=f"{self.live_server_url}/http/getonly/400/")
         uv.check_url()
         self.assertEqual(uv.status, True)
         self.assertEqual(uv.message, '200 OK')
 
     def test_external_check_timedout(self):
-        uv = Url(url="%s/timeout/" % self.live_server_url, still_exists=True)
+        uv = Url(url=f"{self.live_server_url}/timeout/")
         uv.check_url()
         self.assertEqual(uv.status, False)
         self.assertEqual(uv.message, 'Other Error: The read operation timed out')
 
+    def test_working_external_anchor(self):
+        uv = Url(url=f"{self.live_server_url}/http/anchor/#anchor")
+        uv.check_url()
+        self.assertEqual(uv.status, True)
+        self.assertEqual(uv.message, "200 OK, working external hash anchor")
+
+    @patch("linkcheck.models.TOLERATE_BROKEN_ANCHOR", False)
+    def test_broken_external_anchor(self):
+        uv = Url(url=f"{self.live_server_url}/http/anchor/#broken-anchor")
+        uv.check_url()
+        self.assertEqual(uv.status, False)
+        self.assertEqual(uv.message, "200 OK, broken external hash anchor")
+
+    def test_broken_external_anchor_tolerated(self):
+        uv = Url(url=f"{self.live_server_url}/http/anchor/#broken-anchor")
+        uv.check_url()
+        self.assertEqual(uv.status, True)
+        self.assertEqual(uv.message, "200 OK, broken external hash anchor")
+
+    def test_redirect_working_external_anchor(self):
+        uv = Url(url=f"{self.live_server_url}/http/redirect_to_anchor/#anchor")
+        uv.check_url()
+        self.assertEqual(uv.status, True)
+        self.assertEqual(uv.message, "302 Found, working external hash anchor")
+
+    @patch("linkcheck.models.TOLERATE_BROKEN_ANCHOR", False)
+    def test_redirect_broken_external_anchor(self):
+        uv = Url(url=f"{self.live_server_url}/http/redirect_to_anchor/#broken-anchor")
+        uv.check_url()
+        self.assertEqual(uv.status, False)
+        self.assertEqual(uv.message, "302 Found, broken external hash anchor")
+
+    def test_redirect_broken_external_anchor_tolerated(self):
+        uv = Url(url=f"{self.live_server_url}/http/redirect_to_anchor/#broken-anchor")
+        uv.check_url()
+        self.assertEqual(uv.status, True)
+        self.assertEqual(uv.message, "302 Found, broken external hash anchor")
+
+
+class ModelTestCase(TestCase):
+
+    def test_str(self):
+        Author.objects.create(name="John Smith", website="http://www.example.org/smith")
+        self.assertEqual(
+            str(Url.objects.first()),
+            "http://www.example.org/smith",
+        )
+        self.assertEqual(
+            str(Link.objects.first()),
+            "http://www.example.org/smith (Author object (1))",
+        )
+
+    def test_repr(self):
+        Author.objects.create(name="John Smith", website="http://www.example.org/smith")
+        self.assertEqual(
+            repr(Url.objects.first()),
+            "<Url (id: 1, url: http://www.example.org/smith)>",
+        )
+        self.assertEqual(
+            repr(Link.objects.first()),
+            (
+                "<Link (id: 1, url: <Url (id: 1, url: http://www.example.org/smith)>, "
+                "source: <Author: Author object (1)>)>"
+            ),
+        )
+
 
 class ChecklinksTestCase(TestCase):
-    def setUp(self):
-        request.urlopen = mock_urlopen
 
     def test_checklinks_command(self):
         Book.objects.create(title='My Title', description="""
@@ -284,9 +349,30 @@ class FindingLinksTestCase(TestCase):
         self.assertEqual(Url.objects.all().count(), 2)
         self.assertQuerysetEqual(
             Url.objects.all().order_by('url'),
-            ["<Url: http://www.example.org>", "<Url: http://www.example.org/logo.png>"]
+            ["http://www.example.org", "http://www.example.org/logo.png"],
+            transform=lambda obj: obj.url
         )
 
+    def test_urls_exceeding_max_length(self):
+        self.assertEqual(Url.objects.all().count(), 0)
+        with self.assertLogs(logger="linkcheck", level="WARN") as cm:
+            Book.objects.create(
+                title="My Title",
+                description=(
+                    "Here's a link: <a href='http://www.example.org'>Example</a>, and here's a url exceeding "
+                    f"the max length: <img src='http://www.example.org/{MAX_URL_LENGTH * 'X'}' alt='logo'>"
+                ),
+            )
+        # We skip urls which are too long because we can't store them in the database
+        self.assertIn(
+            (
+                "WARNING:linkcheck.listeners:URL exceeding max length will be skipped: "
+                f"http://www.example.org/{MAX_URL_LENGTH * 'X'}"
+            ),
+            cm.output,
+        )
+        self.assertEqual(Url.objects.all().count(), 1)
+
     def test_empty_url_field(self):
         """
         Test that URLField empty content is excluded depending on ignore_empty list.
@@ -304,18 +390,40 @@ class FindingLinksTestCase(TestCase):
         self.assertEqual(Url.objects.all().count(), 2)
 
     def test_findlinks_command(self):
-        all_linklists = apps.get_app_config('linkcheck').all_linklists
-        all_linklists['Authors'].url_fields = []
-        Author.objects.create(name="John Smith", website="http://www.example.org/smith")
-        all_linklists['Authors'].url_fields = ['website']
-
-        out = StringIO()
-        call_command('findlinks', stdout=out)
-        self.assertEqual(
-            out.getvalue(),
-            "Finding all new links...\n"
-            "1 new Url object(s), 1 new Link object(s), 0 Url object(s) deleted\n"
-        )
+        # Disable listeners to only check the management command
+        with disable_listeners():
+            Author.objects.create(name="John Smith", website="https://www.example.org/smith")
+            self.assertEqual(
+                findlinks(),
+                "Updating all links...\n"
+                "Urls: 1 created, 0 deleted, 0 unchanged\n"
+                "Links: 1 created, 0 deleted, 0 unchanged\n"
+            )
+            Author.objects.create(name="John Doe", website="https://www.example.org/doe")
+            Book.objects.create(
+                title='My Title',
+                description="My fav author: <a href='https://www.example.org/doe'>John Doe</a>"
+            )
+            self.assertEqual(
+                findlinks(),
+                "Updating all links...\n"
+                "Urls: 1 created, 0 deleted, 1 unchanged\n"
+                "Links: 2 created, 0 deleted, 1 unchanged\n"
+            )
+            Author.objects.get(name="John Doe").delete()
+            self.assertEqual(
+                findlinks(),
+                "Updating all links...\n"
+                "Urls: 0 created, 0 deleted, 2 unchanged\n"
+                "Links: 0 created, 1 deleted, 2 unchanged\n"
+            )
+            Book.objects.first().delete()
+            self.assertEqual(
+                findlinks(),
+                "Updating all links...\n"
+                "Urls: 0 created, 1 deleted, 1 unchanged\n"
+                "Links: 0 created, 1 deleted, 1 unchanged\n"
+            )
 
 
 class ObjectsUpdateTestCase(TestCase):
@@ -383,17 +491,44 @@ class RegisteringTests(TestCase):
         register_listeners()
 
 
+class QueueTests(TestCase):
+    def test_queue_handling_continue_on_task_crash(self):
+        assert tasks_queue.empty() is True
+
+        def raising():
+            raise RuntimeError("Failing task")
+
+        def passing():
+            pass
+
+        for func in (raising, passing):
+            tasks_queue.put({
+                'target': func,
+                'args': (),
+                'kwargs': {},
+            })
+        with self.assertLogs() as cm:
+            linkcheck_worker(block=False)
+        self.assertEqual(
+            cm.output[0].split('\n')[0],
+            'ERROR:linkcheck.listeners:RuntimeError while running raising with '
+            'args=() and kwargs={}: Failing task'
+        )
+
+
 class ViewTestCase(TestCase):
     def setUp(self):
         User.objects.create_superuser('admin', 'admin@example.org', 'password')
 
     def test_display_url(self):
-        Book.objects.create(title='My Title', description="""Here's a link: <a href="http://www.example.org">Example</a>""")
+        Book.objects.create(
+            title='My Title', description="Here's a link: <a href='http://www.example.org'>Example</a>"
+        )
         Author.objects.create(name="John Smith", website="http://www.example.org#john")
         self.assertEqual(Link.objects.count(), 2)
         self.assertEqual(
-            set([l.display_url for l in Link.objects.all()]),
-            set(['http://www.example.org', 'http://www.example.org#john'])
+            set([link.display_url for link in Link.objects.all()]),
+            set(["http://www.example.org", "http://www.example.org#john"]),
         )
 
     def test_report_view(self):
@@ -407,7 +542,7 @@ class ViewTestCase(TestCase):
         self.assertContains(
             response,
             '<tr><td>sampleapp.Book</td>'
-            '<td style="font-weight: bold;color:green;">True</td>'
+            '<td style="font-weight: bold;color:green;">Yes</td>'
             '<td style="font-weight: bold;color:green;"></td></tr>',
             html=True,
         )
@@ -425,3 +560,32 @@ class FixtureTestCase(TestCase):
 
     def test_fixture(self):
         self.assertEqual(Book.objects.count(), 1)
+
+
+class FilterCallableTestCase(TestCase):
+    def test_filter_callable(self):
+        all_linklists = apps.get_app_config('linkcheck').all_linklists
+        all_linklists['Journals'].html_fields = []
+        Journal.objects.create(title='My Title', description="""
+            My description <a href="http://www.example.org">Example</a>""")
+        Journal.objects.create(title='My Title', version=1, description="""
+            My new description <a href="http://www.example.org">Example</a>""")
+        all_linklists['Journals'].html_fields = ['description']
+        # assert there are two versions of the same journal
+        self.assertEqual(Journal.objects.count(), 2)
+        # assert command just finds the latest version of same journals
+        self.assertEqual(
+            findlinks(),
+            "Updating all links...\n"
+            "Urls: 1 created, 0 deleted, 0 unchanged\n"
+            "Links: 1 created, 0 deleted, 0 unchanged\n"
+        )
+
+
+def findlinks():
+    """
+    Helper function for running the findlinks command and checking its output
+    """
+    out = StringIO()
+    call_command('findlinks', stdout=out)
+    return out.getvalue()
