@@ -5,7 +5,10 @@ from django.views.generic import RedirectView
 
 from linkcheck.tests.sampleapp import views
 
-handler404 = lambda *args, **kwargs: http.HttpResponseNotFound('')
+
+def handler404(*args, **kwargs):
+    return http.HttpResponseNotFound("")
+
 
 urlpatterns = [
     path('admin/linkcheck/', include('linkcheck.urls')),
@@ -16,6 +19,8 @@ urlpatterns = [
     path('http/getonly/<int:code>/', views.http_response_get_only),
     path('http/redirect/<int:code>/', views.http_redirect),
     path('http/redirect_to_404/', views.http_redirect_to_404),
+    path('http/redirect_to_anchor/', views.http_redirect_to_anchor),
     path('http/brokenredirect/', RedirectView.as_view(url='/non-existent/')),
+    path('http/anchor/', views.http_response_with_anchor),
     path('timeout/', views.timeout),
 ]
