@@ -2,9 +2,7 @@ import json
 from itertools import groupby
 from operator import itemgetter
 
-import django
 from django import forms
-from django.conf import settings
 from django.contrib.admin.views.decorators import staff_member_required
 from django.contrib.contenttypes.models import ContentType
 from django.core.exceptions import ObjectDoesNotExist
@@ -12,7 +10,8 @@ from django.core.paginator import Paginator
 from django.http import HttpResponse
 from django.shortcuts import render
 from django.templatetags.static import static
-from django.urls import reverse, NoReverseMatch
+from django.urls import NoReverseMatch, reverse
+from django.utils.translation import gettext as _
 from django.views.decorators.csrf import csrf_exempt
 
 from linkcheck import update_lock
@@ -84,22 +83,22 @@ def report(request):
     qset = Link.objects.order_by('-url__last_checked')
     if link_filter == 'show_valid':
         qset = qset.filter(ignore=False, url__status__exact=True)
-        report_type = 'Good Links'
+        report_type = _('Valid links')
     elif link_filter == 'show_unchecked':
         qset = qset.filter(ignore=False, url__last_checked__exact=None)
-        report_type = 'Untested Links'
+        report_type = _('Untested links')
     elif link_filter == 'ignored':
         qset = qset.filter(ignore=True)
-        report_type = 'Ignored Links'
+        report_type = _('Ignored links')
     else:
         qset = qset.filter(ignore=False, url__status__exact=False)
-        report_type = 'Broken Links'
+        report_type = _('Broken links')
 
     paginated_links = Paginator(qset, RESULTS_PER_PAGE, 0, True)
 
     try:
-        page = int(request.GET.get('page', '1'))
-    except:
+        page = int(request.GET.get("page", "1"))
+    except ValueError:
         page = 0
     # offset = (page - 1) * RESULTS_PER_PAGE
     links = paginated_links.page(page)
@@ -126,13 +125,14 @@ def report(request):
                 admin_url = object.get_admin_url()  # TODO allow method name to be configurable
             except AttributeError:
                 try:
-                    admin_url = reverse('admin:%s_%s_change' % (content_type.app_label, content_type.model), args=[ok])
+                    admin_url = reverse(f'admin:{content_type.app_label}_{content_type.model}_change', args=[ok])
                 except NoReverseMatch:
                     admin_url = None
 
             objects.append({
                 'object': object,
-                'link_list': Link.objects.in_bulk([x['id'] for x in og]).values(),  # Convert values_list back to queryset. Do we need to get values() or do we just need a list of ids?
+                # Convert values_list back to queryset. Do we need to get values() or do we just need a list of ids?
+                'link_list': Link.objects.in_bulk([x['id'] for x in og]).values(),
                 'admin_url': admin_url,
             })
         content_types_list.append({
